/****************************************************************/
/*                                                              */
/*                   -------------------------                  */
/*------------------- 6805 Disassembler, V1.0 ------------------*/
/*                   -------------------------                  */
/*                                                              */
/*                     Author: Mark J. Foster                   */
/*                     Last Update:  2 Feb 87                   */
/*                                                              */
/****************************************************************/

#include "stdio.h"

#define OFS	0x1000		/* Offset to actual run location */

char *bratbl[] =  {"BRA\t","BRN\t","BHI\t","BLS\t",
		   "BCC\t","BCS\t","BNE\t","BEQ\t",
		   "BHCC\t","BHCS\t","BPL\t","BMI\t",
		   "BMC\t","BMS\t","BIL\t","BIH\t" };
char *opctbl1[] = {"NEG","???","???","COM",
		   "LSR","???","ROR","ASR",
		   "LSL","ROL","DEC","???",
		   "INC","TST","???","CLR" };
char *opctbl2[] = {"RTI\t","RTS\t","???\t","SWI\t",
		   "???\t","???\t","???\t","???\t",
		   "???\t","???\t","???\t","???\t",
		   "???\t","DAA\t","STOP\t","WAIT\t" };
char *opctbl3[] = {"???\t","???\t","???\t","???\t",
		   "???\t","???\t","???\t","TAX\t",
		   "CLC\t","SEC\t","CLI\t","SEI\t",
		   "RSP\t","NOP\t","???\t","TXA\t" };
char *opctbl4[] = {"SUB","CMP","SBC","CPX",
		   "AND","BIT","LDA","STA",
		   "EOR","ADC","ORA","ADD",
		   "JMP","JSR","LDX","STX" };

char xfertbl[] = {0x20,0x21,0x80,0x81,0xAC,0xBC,0xCC,0xDC,0xEC,0xFC,0};

char mem[4 * 1024];		/* Memory image being disassembled */
char outbuff[132];		/* Output line buffer */
int lc;				/* Current location counter */
FILE *inf;			/* Input file */
FILE *of;			/* Output file */

main(argc, argv)
    int argc;  char *argv[];
{   int count;
    if (argc != 3) help();
    if (!(inf  = fopen(argv[1], "r"))) error("Can't open input file!");
    if (!(of = fopen(argv[2], "w"))) error("Can't open output file!");
    if ((count = fread(mem, 1, sizeof(mem), inf)) == 0) error("Can't read input file!");
    lc = 0;					/* Reset location counter */
    while (lc < count) {			/* Walk through each byte of code */
	disins();				/* Disassemble an instruction */
	}
    fclose(inf);				/* Close the input file */
    fclose(of);					/* Close the output file */
    } /* main */

disins()
{   char ins;					/* Current instruction byte */
    int slc;					/* Start locptr for instruction */
    int temp;
    slc = lc;					/* Set start-of-ins ptr */
    ins = mem[lc++];				/* Read the current opcode */
    switch(ins >> 4) {				/* DO CASE Opcode class (4 M.S. bits) */
    	case 0:					/* BRSET and BRCLR */
		outs((ins & 0x01) ? "BRCLR\t" : "BRSET\t", of); /* BRAnch on bit */
		outnum((ins >> 1) & 0x07, 1);	/* Write the BIT number */
		outs(",");			/* ...with a comma */
		outnum(mem[lc++], 2);		/* Write the target byte */
		outs(",");
		++lc;				/* Point past last byte */
		temp = mem[lc-1];		/* Get the transfer addr */
		if (temp & 0x80) temp |= 0xFF00;/* sign extend it to a word */
		outnum(temp + lc + OFS, 4);	/* Write the target BRA addr */
		break;
	case 1:					/* BSET and BCLR */
		outs((ins & 0x01) ? "BSET\t" : "BCLR\t", of); /* Bit Set/Reset */
		outnum((ins >> 1) & 0x07, 1);	/* Write the BIT number */
		outs(",");			/* ...with a comma */
		outnum(mem[lc++], 2);		/* Write the target byte */
		break;
	case 2:					/* BRAnch */
		outs(bratbl[ins & 0x0F]);	/* Write the BRAnch instruction name */
		++lc;				/* Point past last byte */
		temp = mem[lc-1];		/* Get the transfer addr */
		if (temp & 0x80) temp |= 0xFF00;/* sign extend it to a word */
		outnum(temp + lc + OFS, 4);	/* Write the target BRA addr */
		break;

	case 3:					/* DIRect byte */
		outs(opctbl1[ins & 0x0F]);	/* Write instruction name */
		outs("\t");			/* ...followed by a TAB */
		outnum(mem[lc++], 2);		/* Write the byte addr */
		break;
	case 4:					/* IMPLied A */
		outs(opctbl1[ins & 0x0F]);	/* Write instruction name */
		outs("A\t");			/* Followed by 'A' */
		break;
	case 5:					/* IMPLied X */
		outs(opctbl1[ins & 0x0F]);	/* Write instruction name */
		outs("X\t");			/* Followed by 'X' */
		break;
	
	case 6:					/* BYTE + X */
	case 7:					/* + X */
		outs(opctbl1[ins & 0x0F]);	/* Write instruction name */
		outs("\t");
		if ((ins >> 4) == 0x06) {	/* Have an indirect offset byte */
		    outnum(mem[lc++], 2);	/* Write the addr byte */
		    }
		outs(",X\t");			/* Write the ,X specifier */
		break;
	case 8:	outs(opctbl2[ins & 0x0F]);	/* Write the ins name */
		break;
	case 9: outs(opctbl3[ins & 0x0F]);	/* More simple one-byte */
		break;
	case 0x0A:				/* Immediate byte */
		if (ins == 0xAD) {		/* Special-cased BSR */
		    outs("BSR\t");		/* Write ins name */
		    ++lc;			/* Point past last byte */
		    outnum(mem[lc-1] + lc + OFS, 4); /* Write the target BSR addr */
		    break;
		    }
	case 0x0B:				/* Direct byte */
		outs(opctbl4[ins & 0x0F]);	/* Write instruction name */
		outs("\t");			/* ...followed by a TAB */
		if ((ins >> 4) == 0x0A) outs("#"); /* Write Immediate flag */
		outnum(mem[lc++], 2);		/* Write the byte addr */
		break;
	case 0x0C:				/* Extended 2-byte */
	case 0x0D:				/* Ext + X */
		outs(opctbl4[ins & 0x0F]);	/* Write instruction name */
		outs("\t");
		outnum((mem[lc] << 8) | mem[lc+1], 4); /* Write the ext addr */
		lc += 2;			/* Point past the address */
		if ((ins >> 4) == 0x0D) {	/* Have a ,X form */
		    outs(",X");			/* Write the ,X specifier */
		    }
		break;
	case 0x0E:				/* BYTE + X */
	case 0x0F:				/* + X */
		outs(opctbl4[ins & 0x0F]);	/* Write instruction name */
		outs("\t");
		if ((ins >> 4) == 0x0E) {	/* Have an indirect offset byte */
		    outnum(mem[lc++], 2);	/* Write the addr byte */
		    }
		outs(",X\t");			/* Write the ,X specifier */
		break;
	}
    outln(slc, lc);
    if (strchr(xfertbl, ins)) fputs("\r\n",of);	/* Put blank line on control xfer */
    } /* disins */

outln(start, end)
{   int i, len;
    fprintf(of, "%04xH ", start + OFS);		/* Write the output address */
    len = end - start;				/* Get length of opcode */
    for (i = 0;  i < len;  ++i) {		/* Display opcodes in hex */
	fprintf(of, "%02x ", mem[start + i]);	/* Display a hex byte */
	}
    while (len < 4) {				/* Pad remaining spaces */
	fputs("   ", of);			/* Pad past unused byte */
	++len;					/* Count this byte posn */
	}
    fputs("\t", of);				/* Write a trailing TAB */
    fputs(outbuff, of);				/* Write disassembled instruction */
    fputs("\r\n", of);				/* Terminate the line */
    outbuff[0] = '\0';				/* Clear output buffer */
    } /* outln */

outnum(value, len)
    int value, len;
{   char buff[80];
    switch(len) {
	case 2: sprintf(buff, "$%02x", value);	/* Write a byte value */
		break;
	case 4: sprintf(buff, "$%04x", value);	/* Write word value */
		break;
	default: sprintf(buff, "%d", value);	/* Write decimal number */
		break;
	}
    strcat(outbuff, buff);			/* Write string to output */
    } /* outnum */

outs(str)
    char *str;
{   strcat(outbuff, str);
    } /* outs */


error(str)
    char *str;
{   puts(str);
    exit(1);
    } /* error */

help()
{   puts("Motorola 6805 Disassembler, V1.0\n");
    puts("Syntax: DISASM05 <infile> <listfile>");
    } /* help */

